/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.export;

import com.vividsolutions.jts.geom.*;
import java.math.BigDecimal;
import java.util.List;
import nl.cloudfarming.client.field.model.TreatmentZone;
import nl.cloudfarming.client.isobus.model.GridTypeEnum;
import nl.cloudfarming.client.isobus.model.IsobusGrid;
import nl.cloudfarming.client.isobus.model.IsobusTreatmentZone;
import nl.cloudfarming.client.lib.geotools.GeometryTools;
import org.geotools.geometry.jts.JTSFactoryFinder;

/**
 * Static class to generate a grid from a collection of treatmentzones and a
 * specified amount of columns and rows.
 *
 * @author Gerben Feenstra
 */
class GridGenerator {

    private static final String EMPTY_TREATMENTZONE = "Empty";

    /**
     * Create a {@link TreatmentZoneGrid} using the passed parameters.
     *
     * Note1: grid y-axis goes from south to north, x-axis goes from west to
     * east. Because y-axis goes from south to north, the order of rows in the
     * returned grid goes from south to north as well. (Position 0 in the List
     * being most southern row in the grid)
     *
     * Note2: TreatmentzoneCode with value 0 is reserved! If no treatmentzone
     * can be found for a gridcell, 0 will be used as the treatmentzoneCode.
     *
     * @param cellWidth the width of the cells used for the grid
     * @param cellHeight the height of the cells used for the grid
     * @param startLongitude the starting longitude of the grid
     * @param startLatitude the starting latitude of the grid
     * @param numberOfRows the number of rows the grid should have
     * @param numberofCols the number of columns the grid should have
     * @param treatmentZones the treatment zones that should be placed in the
     * grid
     * @param matcher the matcher that should be used to determine the geometry
     * for a specific grid cell
     * @return A TreatmentZoneGrid containing all the passed treatment zones
     */
    static TreatmentZoneGrid generate(double cellWidth, double cellHeight,
            double startLongitude, double startLatitude,
            int numberOfRows, int numberofCols,
            List<TreatmentZone> treatmentZones,
            double defaultValue,
            CellMatchingStrategy matcher) {
        // Create empty treatmentzone
        TreatmentZoneGrid grid = new TreatmentZoneGrid(numberOfRows, numberofCols, cellWidth, cellHeight, startLongitude, startLatitude, treatmentZones, defaultValue);

        //latitude needs to be inverted because it starts at the bottom
        startLatitude = startLatitude + (numberOfRows * cellHeight);

        // Row number starting at 0 and working up
        for (int rowNumber = 0; rowNumber < numberOfRows; rowNumber++) {
            // Loop through each column in this row
            for (int columnNumber = 0; columnNumber < numberofCols; columnNumber++) {

                Geometry gridCellGeometry = getCellGeometry(startLongitude, startLatitude, cellWidth, cellHeight, columnNumber, rowNumber);
                TreatmentZone treatmentZone = matcher.findTreatmentZone(treatmentZones, gridCellGeometry);

                if (treatmentZone != null) {
                    grid.insert(rowNumber, columnNumber, grid.getTreatmentZoneCode(treatmentZone), grid.getTreatmentZoneValue(treatmentZone));
                } else {
                    grid.insert(rowNumber, columnNumber, (short) 0, Double.valueOf(defaultValue).longValue());
                }
            }
        }

        return grid;
    }

    /**
     * Create a {@link TreatmentZoneGrid} using the passed parameters.
     *
     * Note1: grid y-axis goes from south to north, x-axis goes from west to
     * east. Because y-axis goes from south to north, the order of rows in the
     * returned grid goes from south to north as well. (Position 0 in the List
     * being most southern row in the grid)
     *
     * Note2: TreatmentzoneCode with value 0 is reserved! If no treatmentzone
     * can be found for a gridcell, 0 will be used as the treatmentzoneCode.
     *
     * @param cellWidth the width of the cells used for the grid
     * @param cellHeight the height of the cells used for the grid
     * @param startLongitude the starting longitude of the grid
     * @param startLatitude the starting latitude of the grid
     * @param numberOfRows the number of rows the grid should have
     * @param numberofCols the number of columns the grid should have
     * @param treatmentZones the treatment zones that should be placed in the
     * grid
     * @return A TreatmentZoneGrid containing all the passed treatment zones
     */
    static TreatmentZoneGrid generate(double cellWidth, double cellHeight,
            double startLongitude, double startLatitude,
            int numberOfRows, int numberofCols,
            List<TreatmentZone> treatmentZones,
            double defaultValue) {
        // use most accurate matcher as default:
        return generate(cellWidth, cellHeight, startLongitude, startLatitude, numberOfRows, numberofCols, 
                treatmentZones, defaultValue, new IntersectionStrategy());
    }

    /**
     * Return a geometry of the specified cell (position), calculated using
     * Startingpoint X, Y as a basis The starting point is the max east and max
     * south position of the containing grid The column and row starting point
     * is assumed to be the top left corner
     *
     * @param startLongitude Startingpoint X
     * @param startLatitude Startingpoint Y
     * @param cellWidth The width of a single cell in the grid
     * @param cellHeight The height of a single cell in the grid
     * @param columnNumber x position in the grid
     * @param rowNumber y position in the grid
     * @return the geometry of the gridcell
     */
    static Geometry getCellGeometry(double startLongitude, double startLatitude, double cellWidth, double cellHeight, int columnNumber, int rowNumber) {

        double cellMinX = startLongitude + (columnNumber * cellWidth);
        double cellMinY = startLatitude - ((rowNumber * cellHeight) + cellHeight);
        double cellMaxX = startLongitude + (columnNumber * cellWidth) + cellWidth;
        double cellMaxY = startLatitude - (rowNumber * cellHeight);

        // Create new geometry
        GeometryFactory factory = JTSFactoryFinder.getGeometryFactory(null);
        Coordinate[] coordinates = new Coordinate[5];

        coordinates[0] = new Coordinate(cellMinX, cellMinY);
        coordinates[1] = new Coordinate(cellMinX, cellMaxY);
        coordinates[2] = new Coordinate(cellMaxX, cellMaxY);
        coordinates[3] = new Coordinate(cellMaxX, cellMinY);
        coordinates[4] = new Coordinate(cellMinX, cellMinY);

        LinearRing ring = factory.createLinearRing(coordinates);
        return factory.createPolygon(ring, null);
    }

    static interface CellMatchingStrategy {

        /**
         * Find the treatmentzone for the specified gridcell. A gridcell can
         * have at most one treatmentzone.
         *
         * @param treatmentZones List of possible treatmentzones
         * @param gridCell Geometry of the gridcell
         * @return the matched treatmentzone, or null when none was found.
         */
        TreatmentZone findTreatmentZone(final List<TreatmentZone> treatmentZones, final Geometry gridCell);
    }

    /**
     * Match a treatmentzone to the cell by testing for intersections. If
     * multiple treatmentzones intersect with the gridcell the treatmentzone
     * which has the largest intersection area with the gridcell will be
     * returned.
     */
    static class IntersectionStrategy implements CellMatchingStrategy {

        @Override
        public TreatmentZone findTreatmentZone(List<TreatmentZone> treatmentZones, Geometry gridCell) {
            // initial maximum of 0 also prevents selection of geometries that only share a boundary.
            double largestIntersectionArea = 0.0;
            TreatmentZone largestIntersectionTZ = null;

            // Find the treatmentzone which has the largest intersection area with the specified gridcell:
            for (TreatmentZone treatmentZone : treatmentZones) {
                Geometry treatmentZoneGeometry = treatmentZone.getGeometry();
                if (gridCell.intersects(treatmentZoneGeometry)) {
                    double intersectionArea = gridCell.intersection(treatmentZoneGeometry).getArea();
                    if (intersectionArea > largestIntersectionArea) {
                        largestIntersectionArea = intersectionArea;
                        largestIntersectionTZ = treatmentZone;
                    }
                }
            }
            return largestIntersectionTZ;
        }
    }

    /**
     * Will create an IsobusGrid object using the passed
     * {@link TreatmentZoneGrid}.
     *
     * @param TreatmentZoneGrid the TreatmentZoneGrid used as the base for this
     * IsobusGrid
     * @param gridType the type of grid that will be created
     * @param gridFileName the filename used for the grid
     * @param treatMentZoneCode only required for the
     * {@link GridTypeEnum#TYPE_2} the code of the only
     * {@link IsobusTreatmentZone} in the task
     * @return the created IsobusGrid
     */
    static IsobusGrid createGrid(TreatmentZoneGrid treatmentZoneGrid, GridTypeEnum gridType, String gridFileName, Short treatMentZoneCode) {
        if (GridTypeEnum.TYPE_2.equals(gridType) && treatMentZoneCode == null) {
            throw new IllegalArgumentException("Grid type 2 needs a treatment zone code");
        }

        IsobusGrid grid = new IsobusGrid();
        grid.setMaximumColumn(treatmentZoneGrid.getNumberOfColumns());
        grid.setMaximumRow(treatmentZoneGrid.getNumberOfRows());
        grid.setCellEastSize(treatmentZoneGrid.getCellWidth());
        grid.setCellNorthSize(treatmentZoneGrid.getCellHeight());
        grid.setMinimumEastPosition(new BigDecimal(treatmentZoneGrid.getLongitude()));
        grid.setMinimumNorthPosition(new BigDecimal(treatmentZoneGrid.getLatitude()));
        grid.setType(gridType);
        grid.setFilename(gridFileName);
        if (GridTypeEnum.TYPE_2.equals(gridType)) {
            grid.setTreatmentZoneCode(treatMentZoneCode);
        }
        return grid;
    }

    /**
     * <p>Create a {@link IsobusGrid}</p> <p>Can create grids for
     * {@link GridTypeEnum#TYPE_1} and {@link GridTypeEnum#TYPE_2}. For type one
     * will set no {@link IsobusGrid#treatmentZoneCode}. For type 2 will set the
     * code of the prototype {@link IsobusTreatmentZone} in the provided
     * list</p>
     *
     * @param treatmentZones if list is empty and {@link GridTypeEnum#TYPE_2} is
     * passed will throw an assertion exception
     * @param grid
     * @param gridType
     * @param gridFileName
     * @return the Created isobusGrid
     */
    static IsobusGrid createIsobusGrid(List<IsobusTreatmentZone> treatmentZones, TreatmentZoneGrid grid, GridTypeEnum gridType, String gridFileName) {
        IsobusGrid isobusGrid = null;
        if (GridTypeEnum.TYPE_1.equals(gridType)) {
            isobusGrid = createGrid(grid, gridType, gridFileName, null);
        } else if (GridTypeEnum.TYPE_2.equals(gridType)) {
            assert !treatmentZones.isEmpty() : "Cannot create a type 2 IsobusGrid without treatmentzones";
            assert treatmentZones.size() >= 3 : "There should be at least 3 treatmentzones (empty, default, and another one)";
            // TZN 0: the empty tzn, TZN 1: the default tzn, TZN 2: the prototype tzn
            isobusGrid = createGrid(grid, gridType, gridFileName, treatmentZones.get(2).getTreatmentZoneCode());
        }
        return isobusGrid;
    }

    /**
     * Match a treatmentzone to the cell by testing if it contains the cells
     * centroid. about twice as fast as the intersection strategy, but less
     * accurate.
     */
    static class CentroidStrategy implements CellMatchingStrategy {

        @Override
        public TreatmentZone findTreatmentZone(List<TreatmentZone> treatmentZones, Geometry gridCell) {
            // Find the treatmentzone which contains the center of the specified gridcell:
            for (TreatmentZone treatmentZone : treatmentZones) {
                if (treatmentZone.getGeometry().contains(gridCell.getCentroid())) {
                    return treatmentZone;
                }
            }
            return null;
        }
    }
}
