/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it is applied to
 * this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with AgroSense.  If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.isobus.export;

import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import java.io.Serializable;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import nl.cloudfarming.client.field.model.TreatmentZone;
import nl.cloudfarming.client.model.AgroURI;
import nl.cloudfarming.client.model.EnterpriseIdProvider;
import org.geotools.coverage.CoverageFactoryFinder;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.geometry.BoundingBox;

/**
 * Class which can be used to store a grid in. This grid contains
 * {@link TreatmentZone}'s
 *
 * @author Merijn Zengers
 * @author Maciek Dulkiewicz
 */
public class TreatmentZoneGrid implements Serializable {

    private TreatmentZoneData[][] grid;
    private int numberOfColumn;
    private int numberOfRows;
    private double cellWidth;
    private double cellHeight;
    private double longitude;
    private double latitude;
    private Map<TreatmentZone, TreatmentZoneData> treatmentZoneCodeMap;
    private static final String EMPTY_TZN_NAME = "empty";
    private static final EnterpriseIdProvider EIP = new AgroURI("AGRONL/1/TZN/0");
    private final TreatmentZone emptyZone = createEmptyTreatmentZone();
    private final double defaultValue;

    public TreatmentZoneGrid(int numberOfRows, int numberOfColumns, double cellWidth, double cellHeight, double longitude, double latitude,
            List<TreatmentZone> treatmentZones, double defaultValue) {
        grid = new TreatmentZoneData[numberOfRows][numberOfColumns];
        this.numberOfRows = numberOfRows;
        this.numberOfColumn = numberOfColumns;
        this.cellWidth = cellWidth;
        this.cellHeight = cellHeight;
        this.longitude = longitude;
        this.latitude = latitude;
        this.defaultValue = defaultValue;

        createTreatmentZoneCodeMap(treatmentZones);
    }

    /**
     * This method creates a map of treatment zones assigning a unique index to
     * each treatment zone passed as a parameter to this method. This method
     * also inserts an empty treatment zone with the index 0 and adds it to the
     * map.
     *
     * @param treatmentZones The treatment zones that should be mapped
     */
    private void createTreatmentZoneCodeMap(List<TreatmentZone> treatmentZones) {
        treatmentZoneCodeMap = new HashMap<>();
        //insert an empty TreatmentZone at the start of this map with index 0
        treatmentZoneCodeMap.put(emptyZone, new TreatmentZoneData((short) 0, Double.valueOf(defaultValue).longValue()));

        short counter = 1;
        for (TreatmentZone treatmentZone : treatmentZones) {
            if (treatmentZone.getValue() != null) {
                treatmentZoneCodeMap.put(treatmentZone, new TreatmentZoneData(counter++, treatmentZone.getValue().longValue()));
            }
        }
    }

    private TreatmentZone createEmptyTreatmentZone() {
        TreatmentZone emptyTreatmentZone = new TreatmentZone(EIP);
        emptyTreatmentZone.setName(EMPTY_TZN_NAME);
        emptyTreatmentZone.setValue(null);
        return emptyTreatmentZone;
    }

    /**
     * Get the code of the TreatMentzone
     *
     * @param treatmentZone
     * @return
     */
    public short getTreatmentZoneCode(TreatmentZone treatmentZone) {
        if (treatmentZone.getValue() == null){
            return treatmentZoneCodeMap.get(emptyZone).getCode();
        }        
        return treatmentZoneCodeMap.get(treatmentZone).getCode();
    }

    /**
     * Get the value of the treatment zone
     *
     * @param treatmentZone
     * @return
     */
    public long getTreatmentZoneValue(TreatmentZone treatmentZone) {
        if (treatmentZone.getValue() == null){
            return treatmentZoneCodeMap.get(emptyZone).getValue();
        }
        return treatmentZoneCodeMap.get(treatmentZone).getValue();
    }

    /**
     * This method searches the treatment zone map for a treatment zone mapped
     * to a specific index.
     *
     * @param code the index of the treatment zone that needs to be found
     * @return the first treatment zone that is found or null if not treatment
     * zone is found at all
     */
    public TreatmentZone getTreatmentZone(int code) {
        for (TreatmentZone treatmentZone : treatmentZoneCodeMap.keySet()) {
            if (treatmentZoneCodeMap.get(treatmentZone).getCode() == code) {
                return treatmentZone;
            }
        }
        return null;
    }

    /**
     * Creates a GridCoverage2D object based on this TreatmentZoneGrid. The
     * resulting object can be displayed by a GridCoverage2DPreviewPanel.
     *
     * @return GridCoverage2D object based on this TreatmentZoneGrid
     */
    public GridCoverage2D getGridCoverage2D() {
        assert (getNumberOfColumns() * getNumberOfRows() == getSize());

        GridCoverage2D coverage;

        double minx = getLongitude();
        double maxx = getLongitude() + (getCellWidth() * getNumberOfColumns());
        double miny = getLatitude();
        double maxy = getLatitude() + (getCellHeight() * getNumberOfRows());

        WritableRaster raster = fillRasterContents();

        BoundingBox boundingBox = new ReferencedEnvelope(minx, maxx, miny, maxy, DefaultGeographicCRS.WGS84);

        GridCoverageFactory factory = CoverageFactoryFinder.getGridCoverageFactory(null);
        coverage = factory.create("name", raster, boundingBox);

        return coverage;
    }

    private WritableRaster fillRasterContents() {
        int rows = getNumberOfRows();
        int columns = getNumberOfColumns();
        int bands = 1;
        Point location = new Point(0, 0);

        WritableRaster raster = WritableRaster.createBandedRaster(DataBuffer.TYPE_INT, columns, rows, bands, location);
        for (int row = 0; row < rows; row++) {
            for (int column = 0; column < columns; column++) {
                int treatmentZoneCode = getRow(row)[column].getCode();
                TreatmentZone treatmentZone = getTreatmentZone(treatmentZoneCode);
                //FIXME: the empty value should be defined in a api. See GridCoverage2DRenderer
                double treatMentZoneColor = treatmentZone.getValue() == null ? Integer.MIN_VALUE : treatmentZone.getValue();
                raster.setSample(column, row, 0, treatMentZoneColor);
            }
        }
        return raster;
    }

    /**
     * Inserts a treatment zone index at the specified row and column Take
     * notice!! this will replace the value current on this row column if there
     * is any
     *
     * @param column
     * @param row
     * @param treatmentZoneIdex
     */
    public void insert(int row, int column, short treatmentZoneIdex, long treatmentZoneValue) {
        grid[row][column] = new TreatmentZoneData(treatmentZoneIdex, treatmentZoneValue);
    }

    /**
     * Gets a row from the grid. Take notice row numbers start at zero
     *
     * @param rowNum
     * @return int array containing treatment zone indices or null if the rowNum
     * does not exist
     */
    public TreatmentZoneData[] getRow(int rowNum) {
        if (grid.length <= rowNum) {
            return null;
        }
        return grid[rowNum];
    }

    public int getNumberOfColumns() {
        return numberOfColumn;
    }

    public int getNumberOfRows() {
        return numberOfRows;
    }

    public double getCellWidth() {
        return cellWidth;
    }

    public double getCellHeight() {
        return cellHeight;
    }

    public double getLatitude() {
        return latitude;
    }

    public double getLongitude() {
        return longitude;
    }

    public int getSize() {
        return numberOfColumn * numberOfRows;
    }

    /**
     * Holds data from treatment zone Holds its code and value
     */
    public class TreatmentZoneData implements Serializable {

        private Short code;
        private long value;

        public TreatmentZoneData(Short code, long value) {
            this.code = code;
            this.value = value;
        }

        public Short getCode() {
            return code;
        }

        public void setCode(Short code) {
            this.code = code;
        }

        public long getValue() {
            return value;
        }

        public void setValue(long value) {
            this.value = value;
        }
    }
}
